/******************************************************************************
			      Voxclock - NetBSD Speaking Clock

    	    Copyright 1993 Martin H. Smith All Rights Reserved.
*******************************************************************************

Project		: VOXCLOCK
Filename	: loader.c
Author		: M.H. Smith (msmith@lssec.bt.co.uk)
Purpose		: Reads in sample files and voice files
Created		: 18th September 1993
RCSid		: $Id: loader.c,v 1.3 1993/09/22 19:31:29 msmith Exp $

Notes:

This program may be freely distributed and copied in source and binary form as
long as it is not sold and the header and copyright information is included in
all such copies.

Every care has been taken in the preparation of this software. It is used at
the recipients risk without charge and I can accept no responsibility for loss
or damage howsoever caused.

******************************************************************************/

#include <stdlib.h>
#include <syslog.h>

#include "voxclock.h"

static word_struct_ptr word_table = NULL;		  /* word sample data */
static int word_base = 0;						  /* awful hacky thing */

String voc_header = "Creative Voice File";		  /* what's a VOC file? */

String words_needed[] =							  /* must have these words */

{
	"1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12",
	"30",
	"its", "oclock", "a.m.", "p.m.",

	NULL
};


/*
 * Voice file was created on my Atari which has a sensible ordering of bytes
 * within a word. Bloody intel
 */

int32 get32(unsigned char *x)

{
	int res = 0;
	int f;

	for (f = 0; f < 4; ++f)
		res = res * 256 + x[f];

	return res;
}


int32 get16(unsigned char *x)

{
	int res = 0;
	int f;

	for (f = 0; f < 2; ++f)
		res = res * 256 + x[f];

	return res;
}


/*
 * Read in a voicebank file. The file format was done on a 68000 so we have
 * to scramble the bytes into the stupid order this silly thing expects!
 */

String read_voicebank(String fname)

{
	Stream fd = fopen(fname, "r");
	String res = NULL;
	
	if (fd)
	{
		int32 flen;
		
		fseek(fd, 0, 2);
		flen = ftell(fd);
		fseek(fd, 0, 0);

		res = malloc(flen);

		if (res)
		{
			fread(res, flen, 1, fd);
		}
		else
			fprintf(stderr, "No memory to load samples\n");
		
		fclose(fd);
	}
	else
		fprintf(stderr, "voxclock: Can't open voicebank file '%s'\n", fname);

	return res;
}


/*
 * Some fields in the voicebank must make sense before we try and process it
 * this is sample rate, file size, file id etc
 */

boolean validate_voicebank(String ptr, int *speed)

{
	boolean res = FALSE;
	
	if (strncmp(ptr, VOX_FILE_ID, strlen(VOX_FILE_ID)))
		fprintf(stderr, "File does not seem to be valid voicebank file\n");
	else
	{
		int stereo = get16(ptr + 12);
		int bits = get16(ptr + 14);
		int rate = get32(ptr + 22) &0xffffff;
		
		if (stereo)
			fprintf(stderr, "Stereo samples not supported, sorry (%d)\n",
					stereo);
		else
		{
			if (bits != 8)
				fprintf(stderr, "Only 8 bit samples supported (%d)\n", bits);
			else
			{
				*speed = rate;
				
				res = TRUE;
			}
		}
	}

	return res;
}



/*
 * Fill in the initial word table, create structures and mark them all as
 * unseen
 */

static void init_word_table(void)

{
	int f = 0;
	word_struct_ptr w;
	int x;

	while (words_needed[f])
		++f;

	w = (word_struct_ptr) malloc(sizeof(word_struct) * (f + 1));

	if (w)
	{
		for (x = 0; x <= f; ++x)
		{
			w[x].name = words_needed[x];		  /* copy name of word */
			w[x].found = FALSE;					  /* not seen it */
			w[x].ptr = NULL;					  /* have no pointer to it */
			w[x].length = 0;
		}

		word_table = w;
	}
	else
		fprintf(stderr, "voxclock: No memory for word table\n");
}


/*
 * Given a string, look it up in the word table
 */

static word_struct_ptr find_word(String name)

{
	int f = 0;
	word_struct_ptr res = NULL;

	for (f = 0; word_table[f].name && !res; ++f)
		if (!strcmp(name, word_table[f].name))
			res = &word_table[f];
	
	return res;
}


/*
 * Look through the word list in the file, work out how many words we have
 * and build appropriate data structures for them that match the ones
 * we are expecting. It is an error if we can't find all the words we want
 * or there are duplicate words
 */

boolean extract_words(String ptr)

{
	boolean res = FALSE;
	int f;

	String oldptr = ptr;
	
	init_word_table();
	ptr += 128;									  /* skip header */
	
	while (*ptr)
	{
		word_struct_ptr w;
		
		w = find_word(ptr);

		if (!w)
			fprintf(stderr, "voxclock: Unknown word '%s'\n", ptr);
		else
		{
			if (w -> found)
				fprintf(stderr, "voxclock: Duplicate word '%s' ignored\n",
						ptr);
			else
			{
				w -> found = TRUE;
				w -> length = get32(ptr + 16);
				w -> ptr = (String) (get32(ptr + 12) + (int) oldptr); 
			}
		}
		
		ptr += 20;
	}

	word_base = ptr - oldptr;

	res = TRUE;
	
	for (f = 0; word_table[f].name; ++f)
	{
		if (!word_table[f].found)
		{
			fprintf(stderr, "voxclock: Word %s was not found in voicebank\n",
					word_table[f].name);
			res = FALSE;
		}
	}
	
	return res;
}


/*
 * Test whether a file is a Creative Labs .VOC file or not
 */

boolean is_voc_file(String ptr, int *offset)

{
	boolean res = (strncmp(ptr, voc_header,
						   strlen(voc_header))) ? FALSE : TRUE;

	if (res)									  /* looks ok so far */
	{
		int vers;
		int vers2;
		*offset = ptr[20] + (ptr[21] * 256);

		vers  = ptr[22] + (ptr[23] * 256);
		vers2 = ptr[24] + (ptr[25] * 256);

		if (((~(vers) + 0x1234) & 0xffff) != vers2)
		{
			syslog(LOG_ERR, "Version check mismatch - bad .VOC format");
			res = FALSE;
		}
	}

	return res;
}


/*
 * Read in a sample file
 */

String read_sample_file(String name, int *len)

{
	String res = NULL;
	Stream fd = fopen(name, "r");

	if (fd)
	{
		int32 flen;
		
		fseek(fd, 0, 2);
		flen = ftell(fd);
		fseek(fd, 0, 0);

		*len = flen;
		
		res = malloc(flen);
		
		if (res)
			fread(res, flen, 1, fd);
		else
			fprintf(stderr, "No memory to load sample\n");
		
		fclose(fd);
	}
	else
		fprintf(stderr, "voxclock: Can't open sample file '%s'\n", name);

	return res;
}


/*
 * We know it's a VOC file, extract the sample information from it and
 * set up the pointers. Should really handle multiple blocks and stuff but
 * for now expects a single one.
 */

int setup_voc_block(unsigned char * start, int *len, int *speed)

{
	int res = 0;
	
	if (*start == 1)							  /* type 1 = voice block */
	{
		int length = start[1] + (256 * start[2]) + (65536 * start[3]);
		int rate = -1000000 / (start[4] - 256);

		res = 4;								  /* skip header */

		*len = length;
		*speed = rate;
	}

	return res;
}


/*
 * Find out the start and length of a word sample
 */

String get_sample_data(word_enum w, int *length)

{
	*length = word_table[w].length;

	return word_table[w].ptr + word_base;
}

/* file identification */

static char RCSid[] = "$Id: loader.c,v 1.3 1993/09/22 19:31:29 msmith Exp $";
