/******************************************************************************
  			        Voxclock - NetBSD Speaking Clock

    	    Copyright 1993 Martin H. Smith All Rights Reserved.
*******************************************************************************

Project		: VOXCLOCK
Filename	: main.c
Author		: M.H. Smith (msmith@lssec.bt.co.uk)
Purpose		: Main program
Created		: 18th September 1993
RCSid		: $Id: main.c,v 1.4 1993/09/22 19:31:29 msmith Exp $

Notes:

This program may be freely distributed and copied in source and binary form as
long as it is not sold and the header and copyright information is included in
all such copies.

Every care has been taken in the preparation of this software. It is used at
the recipients risk without charge and I can accept no responsibility for loss
or damage howsoever caused.

******************************************************************************/

#include <stdio.h>
#include <syslog.h>

#include "voxclock.h"

boolean want_ampm = TRUE;						  /* say 11 AM or 11 o'clk */
boolean half_hour = TRUE;						  /* half hours? */
boolean chime_all = TRUE;						  /* announce or chime ? */
boolean speaking  = TRUE;						  /* announce time ? */
boolean test_mode = FALSE;						  /* chime very soon to test */

int chime_play_rate = 0;						  /* chime sample speed */
int chime_data_len = 0;							  /* how long is chime */
unsigned char * chime_data_ptr = NULL;			  /* where is data for it? */

int voice_play_rate = 0;						  /* speed to play voice at */

String vox_file_name = DEFAULT_VOICE;			  /* standard voice file */
String chime_file_name = DEFAULT_CHIME;			  /* standard chime file */

/*
 * Go through the options and work out what is wanted. Sets up various flags
 * and variables to be used by the program.
 */

static boolean process_options(int argc, String argv[])

{
	int arg = 1;
	int files = 0;
	boolean res = TRUE;
	
	while (arg < argc)							  /* more arguments */
	{
		if (argv[arg][0] == '-')				  /* an option */
		{
			switch (argv[arg][1])
			{
			case 'o':
				want_ampm = FALSE;				  /* say o'clock */
				break;
				
			case 'a':
				want_ampm = TRUE;				  /* say am or pm */
				break;
				
			case '3':
				half_hour = TRUE;				  /* say half hours */
				break;
				
			case '0':
				half_hour = FALSE;				  /* don't say half hour */
				break;
				
			case 'v':
				speaking = TRUE;				  /* say the time */
				break;
				
			case 'n':
				speaking = FALSE;				  /* don't say the time */
				break;
				
			case 'c':
				chime_all = TRUE;				  /* chime all hours */
				break;
				
			case '1':
				chime_all = FALSE;				  /* chime once */
				break;
				
			case 's':
				if (isdigit(argv[arg][2]))
					chime_play_rate = atoi(argv[arg] + 2);
				else
				{
					if (arg == argc)
						fprintf(stderr, "-s option needs sample rate\n");
					else
					{
						++arg;
						chime_play_rate = atoi(argv[arg]);
					}
				}
				break;

			case 't':							  /* select test mode */
				test_mode = TRUE;
				break;
				
			default:
				
				fprintf(stderr, "Unknown option '%s'\n", argv[arg]);
				res = FALSE;
				break;
			}
		}
		else									  /* a filename */
		{
			if (files == 0)
				chime_file_name = argv[arg];
			else
				vox_file_name = argv[arg];
			
			++files;
		}
		
		++arg;
	}
	
	return res;
}


/*
 * Print out a message giving the usage options
 */

static void give_out_usage(void)
	
{
	fprintf(stderr, "VOXCLOCK %s by Martin H. Smith. Built on %s\n\n",
			VOX_VERSION, __DATE__);
	fprintf(stderr, 
			"Usage voxclock [options] [chime_sample_file] [voice_file]\n\n");
	fprintf(stderr, "Options are:\n\n");
	fprintf(stderr, "-o         Announce time as e.g. 11 o'clock\n");
	fprintf(stderr, "-a         Announce time as e.g. 11 A.M. (default)\n");
	fprintf(stderr, "-0         Chime / announce only on the hour\n");
	fprintf(stderr, "-3         Chime / announce half hours too (default)\n");
	fprintf(stderr, "-n         Do not speak the time, chime only\n");
	fprintf(stderr, "-v         Speak the time as well as chime (default)\n");
	fprintf(stderr, "-1         Chime once on the hour\n");
	fprintf(stderr, "-c         Chime the hours appropriately (default)\n");
	fprintf(stderr, "-s <speed> Specify chime playback speed\n\n");
	fprintf(stderr, "Note: you do not need -s if chime is a .VOC file\n");
	fprintf(stderr, 
			"Voice file should be sample file bank built by makebank\n");
}


/*
 * This is where we came in I believe ...
 */

int main(int argc, String argv[])

{
	int exit_status = 0;						  /* all ok? */
	int offset;
	int clength;								  /* chime sample length */

	openlog("voxclock", LOG_PID, LOG_USER);		  /* set up syslog */
	
	if (process_options(argc, argv))			  /* options ok ? */
	{
		String ptr = read_voicebank(vox_file_name);
		String ptr2 = read_sample_file(chime_file_name, &clength);
		
		if (!ptr || !validate_voicebank(ptr, &voice_play_rate))
			exit_status = 1;					  /* rubbishy file */
		else
			if (!extract_words(ptr))			  /* find words */
				exit_status = 1;
		
		if (!exit_status && ptr2 && is_voc_file(ptr2, &offset))
		{
			chime_data_ptr = ptr2 + setup_voc_block(ptr2 + offset,
													&chime_data_len,
													&chime_play_rate);
			chime_data_ptr += offset;
		}
		else										  /* use all of file raw */
		{
			chime_data_ptr = ptr2;
			chime_data_len = clength;
		}
		
		if (chime_data_ptr && chime_data_len && chime_play_rate &&
			voice_play_rate)
		{
			clock_go();
		}
		else
			fprintf(stderr, "voxclock: Initialisation failed. Exiting.\n");
	}
	else
	{
		exit_status = 1;
		give_out_usage();
	}
	
	return exit_status;
}
	
static char RCSid[] = "$Id: main.c,v 1.4 1993/09/22 19:31:29 msmith Exp $";
